@@ -1,21 +1,34 @@
 import pkg_resources
 import argparse
 import logging
-import pushy
+import textwrap
 import sys
+from string import join
 
+import ceph_deploy
 from . import exc
 from . import validate
-from . import sudo_pushy
 from .util import log
 from .util.decorators import catches
 
 LOG = logging.getLogger(__name__)
 
 
-def parse_args(args=None, namespace=None):
+__header__ = textwrap.dedent("""
+    -^-
+   /   \\
+   |O o|  ceph-deploy v%s
+   ).-.(
+  '/|||\`
+  | '|` |
+    '|`
+""" % ceph_deploy.__version__)
+
+
+def get_parser():
     parser = argparse.ArgumentParser(
-        description='Deploy Ceph',
+        formatter_class=argparse.RawDescriptionHelpFormatter,
+        description='Easy Ceph deployment\n\n%s' % __header__,
         )
     verbosity = parser.add_mutually_exclusive_group(required=False)
     verbosity.add_argument(
@@ -29,9 +42,14 @@ def parse_args(args=None, namespace=None):
         help='be less verbose',
         )
     parser.add_argument(
-        '-n', '--dry-run',
-        action='store_true', dest='dry_run',
-        help='do not perform any action, but report what would be done',
+        '--version',
+        action='version',
+        version='%s' % ceph_deploy.__version__,
+        help='the current installed version of ceph-deploy',
+        )
+    parser.add_argument(
+        '--username',
+        help='the username to connect to the remote host',
         )
     parser.add_argument(
         '--overwrite-conf',
@@ -69,20 +87,20 @@ def parse_args(args=None, namespace=None):
     parser.set_defaults(
         # we want to hold on to this, for later
         prog=parser.prog,
-
-        # unit tests can override this to mock pushy; no user-visible
-        # option sets this
-        pushy=pushy.connect,
-
         cluster='ceph',
         )
-    args = parser.parse_args(args=args, namespace=namespace)
-    return args
+    return parser
 
 
-@catches((KeyboardInterrupt, RuntimeError))
+@catches((KeyboardInterrupt, RuntimeError, exc.DeployError,))
 def main(args=None, namespace=None):
-    args = parse_args(args=args, namespace=namespace)
+    parser = get_parser()
+
+    if len(sys.argv) < 2:
+        parser.print_help()
+        sys.exit()
+    else:
+        args = parser.parse_args(args=args, namespace=namespace)
 
     console_loglevel = logging.DEBUG  # start at DEBUG for now
     if args.quiet:
@@ -110,13 +128,6 @@ def main(args=None, namespace=None):
     root_logger.addHandler(sh)
     root_logger.addHandler(fh)
 
-    sudo_pushy.patch()
-
-    try:
-        return args.func(args)
-    except exc.DeployError as e:
-        print >> sys.stderr, '{prog}: {msg}'.format(
-            prog=args.prog,
-            msg=e,
-            )
-        sys.exit(1)
+    LOG.info("Invoked (%s): %s" %(ceph_deploy.__version__,
+                                  join(sys.argv, " ")))
+    return args.func(args)
